#!/usr/bin/perl

##############################################################
# Utility for checking the integrity of a CAS file containing
# a SYSTEM or a BASIC program.
#
# Verson: 0.01
#
# Written by Csaba Markus <csaba.markus.at.freemail.hu>
# September 2005, Budapest.
##############################################################

my @tokens = (
    "END"   ,"FOR"   ,"RESET" ,"SET"   ,"CLS"   ,"CMD"   ,"RANDOM","NEXT"   
   ,"DATA"  ,"INPUT" ,"DIM"   ,"READ"  ,"LET"   ,"GOTO"  ,"RUN"   ,"IF"     
   ,"RESTORE","GOSUB","RETURN","REM"   ,"STOP"  ,"ELSE"  ,"TRON"  ,"TROFF"  
   ,"DEFSTR","DEFINT","DEFSNG","DEFDBL","LINE"  ,"EDIT"  ,"ERROR" ,"RESUME" 
   ,"OUT"   ,"ON"    ,"OPEN"  ,"FIELD" ,"GET"   ,"PUT"   ,"CLOSE" ,"LOAD"   
   ,"MERGE" ,"NAME"  ,"KILL"  ,"LSET"  ,"RSET"  ,"SAVE"  ,"SYSTEM","LPRINT" 
   ,"DEF"   ,"POKE"  ,"PRINT" ,"CONT"  ,"LIST"  ,"LLIST" ,"DELETE","AUTO"   
   ,"CLEAR" ,"CLOAD" ,"CSAVE" ,"NEW"   ,"TAB("  ,"TO"    ,"FN"    ,"USING"  
   ,"VARPTR","USR"   ,"ERL"   ,"ERR"  ,"STRING\$","INSTR","POINT" ,"TIME\$" 
   ,"MEM"   ,"INKEY\$","THEN" ,"NOT"   ,"STEP"  ,"+"     ,"-"     ,"*"      
   ,"/"     ,"["     ,"AND"   ,"OR"    ,">"     ,"="     ,"<"     ,"SGN"    
   ,"INT"   ,"ABS"   ,"FRE"   ,"INP"   ,"POS"   ,"SQR"   ,"RND"   ,"LOG"    
   ,"EXP"   ,"COS"   ,"SIN"   ,"TAN"   ,"ATN"   ,"PEEK"  ,"CVI"   ,"CVS"    
   ,"CVD"   ,"EOF"   ,"LOC"   ,"LOF"   ,"MKI\$" ,"MKS\$" ,"MKD\$" ,"CINT"   
   ,"CSNG"  ,"CDBL"  ,"FIX"   ,"LEN"   ,"STR\$" ,"VAL"   ,"ASC"   ,"CHR\$"  
   ,"LEFT\$","RIGHT\$","MID\$","'"     ,"((fc))","((fd))","((fe))","((ff))" 
);

sub loadfile {
    while (<>) {
        $data .= $_;
    }
    $datap = 0;
    $datal = length($data);
} 

sub getbyte {
    if ($datap >= $datal) {
        printf "Oops... unexpected end of file at position %d, exiting\n", $datap;
        exit;
    }
    return ord(substr($data, $datap++, 1));
}

sub getposition() {
    return $datap;
}

sub getlength() {
    return $datal;
}

sub getstring {
    my $l = $_[0];
    if ($datap + $l > $datal) {
        print "Oops... unexpected end of file";
        exit;
    }
    my $s = substr($data, $datap, $l);
    $datap += $l;
    return $s;
}

sub printblock {
    my $address = $_[0];
    my $octetstr = $_[1];
    my $prefix = $_[2];
    my $BYTESPERLINE = 16;
    while (length($octetstr) > 0) {
        my $l = length($octetstr) <= $BYTESPERLINE ?
            length($octetstr) : $BYTESPERLINE;
        my $i;
        my $dataasstring = "";
        printf "%s%04x", $prefix, $address;
        for ($i = 0; $i < $BYTESPERLINE; $i++) {
            my $code = ord(substr($octetstr, $i, 1));
            $dataasstring .= $i >= $l ?
                " " :
                (($code < 32 || ($code >= 127 && $code < 160))?
                 "." : chr($code));
            if ($i > 0 && ($i & 3) == 0) {
                print $i < $l ? " |" : "  ";
            }
            if ($i < $l) {
                printf " %02x", ord(substr($octetstr, $i, 1));
            }
            else {
                print "   ";
            }
        }
        print "  $dataasstring\n";
        $octetstr = substr($octetstr, $l);
        $address += $l;
    }
}

sub checkeof {
    return if $displayonly;
    my $p = getposition();
    my $l = getlength();
    printf "end of data at position %d\n", $p;
    if ($p < $l) {
        printf "Extra bytes remaining: %d\n", $l - $p;
    }
    else {
        print "end of file\n";
    }
}

sub detokenize {
    my $token = $_[0];
    
    return $token >= 128 ? $tokens[$token - 128] : chr($token);
}

while (scalar @ARGV > 0 &&
       ($ARGV[0] eq "-v" ||
        $ARGV[0] eq "-d" ||
        $ARGV[0] eq "-h")) {
    $verbose = 1 if $ARGV[0] eq "-v";
    $displayonly = 1 if $ARGV[0] eq "-d";
    if ($ARGV[0] eq "-h") {
        print
            "Utility for checking the integrity of a CAS file containing\n" .
            "a SYSTEM or a BASIC program.\n" .
            "Usage: checkcas.pl [options] [filename]\n" .
            "options: -v  verbose: display the contents of the program, too\n" .
            "         -d  display the contents of the program only\n" .
            "         -h  print this help text\n" .
            "If filename is omitted, standard input is assumed.\n" .
            "Version: 0.01\n";
        exit;
    }
    shift @ARGV;
}

loadfile();
while (getbyte() != 0xa5) {}
$b = getbyte();
if ($b == 0x55) {
    $progname=getstring(6);
    print "SYSTEM program \"$progname\"\n"
        unless $displayonly;
    while (($b = getbyte()) == 0x3c) {
        $blocklen = getbyte();
        $blocklen = 256 if $blocklen == 0;
        $loadaddrlow = getbyte();
        $loadaddrhigh = getbyte();
        $loadaddr = $loadaddrlow + 256 * $loadaddrhigh;
        $checksum = $loadaddrlow + $loadaddrhigh;
        $blockstartoffset = getposition();
        $block = "";
        printf "block: %3d bytes at %04XH: ", $blocklen, $loadaddr
            unless $displayonly;
        for ($i = 0; $i < $blocklen; $i++) {
            $databyte = getbyte();
            $checksum += $databyte;
            $block .= chr($databyte);
        }
        $checksuminfile = getbyte();
        $checksum &= 0xff;
        if ($checksum == $checksuminfile) {
            print "ok\n" unless $displayonly;
        }
        else {
            printf "checksum error: %02XH-%02XH (data offset: 0x%04X)\n",
                $checksum, $checksuminfile, $blockstartoffset;
        }
        if ($verbose || $displayonly || $checksum != $checksuminfile) {
            printblock($loadaddr, $block, "  ");
        }
    }
    if ($b == 0x78) {
        $startaddrlow = getbyte();
        $startaddrhigh = getbyte();
        $startaddr = $startaddrlow + 256 * $startaddrhigh;
        printf "start address: %04XH\n", $startaddr;
        checkeof();
    }
    else {
        printf "unknown control byte %02XH at position %04XH, exiting...\n",
        $b, getposition() - 1;
    }
}
elsif ($b == 0xd3) {
    $b = getbyte();
    if ($b == 0xd3) {
        $b = getbyte();
        if ($b == 0xd3) {
            $progname=getstring(1);
            print "BASIC program \"$progname\"\n" unless $displayonly;
            my $nextaddr = -1;
            my $lastoffset = 99999;
            my $notfirstline = 0;
            my $lastlinenum = -1;
            while (1) {
                my $positionerror = "";
                my $linenumerror = "";
                my $colonerror = "";
                my $remarkerror = "";
                my $filepos = getposition();
                my $addr = $nextaddr;
                if ($notfirstline) {
                    my $offset = $addr - $filepos;
                    if ($lastoffset != 99999) {
                        if ($offset != $lastoffset) {
                            $positionerror = sprintf ", offset error: %d",
                                                     $offset - $lastoffset;
                        }
                    }
                    $lastoffset = $offset;
                }
                my $nextaddrlow = getbyte();
                my $nextaddrhigh = getbyte();
                $nextaddr = $nextaddrlow + 256 * $nextaddrhigh;
                my $linenum;
                my $line = "";
                if ($nextaddr == 0) {
                    printf "end of program, last byte at %04XH$positionerror\n",
                           $addr + 1 unless $displayonly;
                    checkeof();
                    last;
                }
                
                my $linenumlow = getbyte();
                my $linenumhigh = getbyte();
                $linenum = $linenumlow + 256 * $linenumhigh;
                if ($lastlinenum >= 0) {
                    if ($linenum <= $lastlinenum) {
                        $linenumerror = " (line number error)";
                    }
                }
                $lastlinenum = $linenum;
                do {
                    $b = getbyte();
                    if ($b != 0) {
                        if ($b == 0x95) {
                            if (length($line) > 0 &&
                                substr($line, -1) eq ":") {
                                $line = substr($line, 0, length($line) - 1);
                            }
                            else {
                                $colonerror = " ELSE token error";
                            }
                        }
                        elsif ($b == 0xfb) {
                            if (length($line) > 3 &&
                                substr($line, -4) eq ":REM") {
                                $line = substr($line, 0, length($line) - 4);
                            }
                            else {
                                $remarkerror = " REMARK token error";
                            }
                        }
                        $line .= detokenize($b);
                    }
                } while ($b != 0);
                my $lineat = $notfirstline ? (sprintf "at %04XH", $addr) : "(first line)";
                if (!$displayonly || length("$linenumerror$positionerror$colonerror$remarkerror") > 0) {
                    print "line $linenum$linenumerror $lineat$positionerror$colonerror$remarkerror\n";
                }
                if ($verbose || $displayonly) {
                    print "$linenum $line\n";
                }
                $notfirstline = 1;
            }
        }
        else {
            printf "Not quite a BASIC program " .
                "(control sequence D3H D3H %02XH), exiting...\n", $b;
        }
    }
    else {
        printf "Not quite a BASIC program " .
            "(control sequence D3H %02XH), exiting...\n", $b;
    }
}
else {
    printf "Not a SYSTEM or BASIC program (control byte %02XH), exiting...\n", $b;
}
